#
# compact_database_dlg.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component
from mnemosyne.pyqt_ui.qt_worker_thread import \
    QtWorkerThread, QtGuiThread
from mnemosyne.pyqt_ui.ui_compact_database_dlg import \
    Ui_CompactDatabaseDlg
from mnemosyne.pyqt_ui.delete_unused_media_files_dlg import \
    DeleteUnusedMediaFilesDlg
from mnemosyne.libmnemosyne.ui_components.dialogs import \
    CompactDatabaseDialog


class CompactThread(QtWorkerThread):

    def __init__(self, defragment_database, archive_old_logs, **kwds):
        super().__init__(**kwds)
        self.defragment_database = defragment_database
        self.archive_old_logs = archive_old_logs

    def do_work(self):
        if self.archive_old_logs:
            self.mnemosyne.database().archive_old_logs()
        if self.defragment_database:
            self.mnemosyne.database().defragment()


class CompactDatabaseDlg(QtWidgets.QDialog, QtGuiThread, CompactDatabaseDialog,
                         Ui_CompactDatabaseDlg):

    def __init__(self, started_automatically=False, **kwds):
        super().__init__(**kwds)
        self.setupUi(self)
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        self.started_automatically = started_automatically

    def activate(self):
        CompactDatabaseDialog.activate(self)
        self.exec_()

    def accept(self):
        defragment_database = \
            (self.defragment_database.checkState() == QtCore.Qt.Checked)
        delete_unused_media_files = \
           (self.delete_unused_media_files.checkState() == QtCore.Qt.Checked)
        archive_old_logs = \
            (self.archive_old_logs.checkState() == QtCore.Qt.Checked)
        if not (defragment_database or delete_unused_media_files or \
                archive_old_logs):
            QtWidgets.QDialog.accept(self)
        if delete_unused_media_files:
            unused_media_files = self.database().unused_media_files()
            if len(unused_media_files) != 0:
                DeleteUnusedMediaFilesDlg(\
                    component_manager=self.component_manager,
                    unused_media_files=unused_media_files).activate()
        if defragment_database or archive_old_logs:
            self.compact_in_thread(defragment_database, archive_old_logs)
        else:
            QtWidgets.QDialog.accept(self)

    def compact_in_thread(self, defragment_database, archive_old_logs):
        self.main_widget().set_progress_text(_("Compacting database..."))
        self.worker_thread = CompactThread(\
            defragment_database, archive_old_logs, mnemosyne=self)
        self.run_worker_thread()

    def work_ended(self):
        self.main_widget().close_progress()
        if not self.started_automatically:
            self.main_widget().show_information(_("Done!"))
            QtWidgets.QDialog.accept(self)


class PyQtDatabaseMaintenance(Component):

    component_type = "database_maintenance"

    def run(self):
        CompactDatabaseDlg(started_automatically=True,
                           component_manager=self.component_manager).\
            compact_in_thread(defragment_database=True, archive_old_logs=True)




#
# mnemosyne2_db.py <Peter.Bienstman@UGent.be>
#

import os

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.utils import expand_path
from mnemosyne.libmnemosyne.file_format import FileFormat
from mnemosyne.libmnemosyne.file_formats.mnemosyne2_cards import Mnemosyne2Cards


class Mnemosyne2Db(FileFormat):

    description = _("Mnemosyne 2.x *.db files")
    extension = ".db"
    filename_filter = _("Mnemosyne 2.x database for merging (*.db)")
    import_possible = True
    export_possible = False

    def do_import(self, filename, extra_tag_names=""):
        db = self.database()
        if filename.endswith("config.db"):
            self.main_widget().show_error(\
                _("The configuration database is not used to store cards."))
            return
        data_dir = self.config().data_dir
        receiving_database_filename = \
            expand_path(self.config()["last_database"], data_dir)
        db.dump_to_science_log()
        # Heuristic to check if we haven't imported this database before.
        current_tag_ids = set([tag.id for tag in db.tags()])
        self.log().active = False
        db.load(filename)
        self.log().active = True
        tag_ids_to_import = set([tag.id for tag in db.tags()])
        if len(tag_ids_to_import.intersection(current_tag_ids)) >= 2:
            answer = self.main_widget().show_question(\
_("It looks like you've imported this database before. Importing it twice will generate duplicate log entries, which will skew your statistics. Do you want to continue?"),
_("Abort"), _("Continue"), "")
            if answer == 0:
                db.load(receiving_database_filename)
                self.log().active = True
                return
        # Export to temporary *.cards file.
        cards_format = Mnemosyne2Cards(self.component_manager)
        tmp_cards_filename = os.path.join(data_dir, "TMP.cards")
        cards_format.do_export(tmp_cards_filename, used_for_merging_dbs=True)
        old_deactivated_card_type_fact_view_ids = \
            db.current_criterion().deactivated_card_type_fact_view_ids
        user_card_types = [card_type for card_type in db.card_types_in_use() \
            if db.is_user_card_type(card_type)]
        # Get config info to be merged.
        old_config_dir = self.config().config_dir
        self.config().config_dir = os.path.dirname(filename)
        self.log().active = False
        self.config().load()
        config_to_merge = self.config().copy()
        self.config().config_dir = old_config_dir
        self.config().load()
        self.log().active = True
        # Import the *.cards file into the receiving database.
        db.load(receiving_database_filename)
        log_index_before_import = db.current_log_index()
        db.importing_with_learning_data = True
        cards_format.do_import(\
            tmp_cards_filename, extra_tag_names, show_metadata=False)
        db.importing_with_learning_data = False
        db.merge_logs_from_other_database(filename, log_index_before_import)
        os.remove(tmp_cards_filename)
        db.current_criterion().deactivated_card_type_fact_view_ids.update(\
            old_deactivated_card_type_fact_view_ids)
        db.set_current_criterion(db.current_criterion())
        for property_name in ["background_colour", "font", "font_colour",
            "alignment", "hide_pronunciation_field"]:
            self.log().edited_setting(property_name)
            for card_type in user_card_types:
                if card_type.id in config_to_merge[property_name]:
                    self.config()[property_name][card_type.id] = \
                        config_to_merge[property_name][card_type.id]
        db.skip_science_log()


#
# review_wdgt.py <Peter.Bienstman@UGent.be>
#

import os
import sys

from PyQt5 import QtCore, QtGui, QtWidgets, QtWebEngineWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.pyqt_ui.ui_review_wdgt import Ui_ReviewWdgt
from mnemosyne.libmnemosyne.ui_components.review_widget import ReviewWidget


class QAOptimalSplit(object):

    """Algorithm to make sure the split between question and answer boxes is
    as optimal as possible.

    This code is not part of ReviewWidget, in order to be able to reuse this
    in Preview Cards.

    """

    used_for_reviewing = True

    def __init__(self, **kwds):
        super().__init__(**kwds)

    def setup(self):
        self.question.settings().setAttribute(\
            QtWebEngineWidgets.QWebEngineSettings.PluginsEnabled, True)
        self.answer.settings().setAttribute(\
            QtWebEngineWidgets.QWebEngineSettings.PluginsEnabled, True)
        # Add some dummy QWebEngineEngineViews that will be used to determine the actual
        # size of the rendered html. This information will then be used to
        # determine the optimal split between the question and the answer
        # pane.
        self.question_preview = QtWebEngineWidgets.QWebEngineView()
        self.question_preview.loadFinished.connect(\
            self.question_preview_load_finished)
        self.answer_preview = QtWebEngineWidgets.QWebEngineView()
        self.answer_preview.loadFinished.connect(\
            self.answer_preview_load_finished)
        # Calculate an offset to use in the stretching factor of the boxes,
        # e.g. question_box = question_label + question.
        self.stretch_offset = self.question_label.size().height()
        if self.question_box.spacing() != -1:
            self.stretch_offset += self.question_box.spacing()
        self.scrollbar_width = QtWidgets.QScrollBar().sizeHint().width()
        self.question_text = ""
        self.answer_text = ""
        self.required_question_size = self.question.size()
        self.required_answer_size = self.answer.size()
        self.is_answer_showing = False
        self.times_resized = 0

    #
    # Code to determine optimal QA split based on prerendering the html
    # in a headless server. Does not work yet in PyQt5, because WebEngine
    # does not yet support headless mode.
    # See: https://bugreports.qt.io/browse/QTBUG-50523
    #

    def resizeEvent_off(self, event):
        # Update stretch factors when changing the size of the window.
        self.set_question(self.question_text)
        self.set_answer(self.answer_text)
        # To get this working the first time we start the program (in general,
        # the first one or two resize events, depending on whether or not we
        # changed the window size), we need to explicitly show the contents.
        # (Qt bug?). We don't do this on subsequent resizes to prevent flicker
        # and media replays.
        if self.times_resized < 2:
            self.reveal_question()
            if self.is_answer_showing:
                self.reveal_answer()
            self.times_resized += 1
        return QtWidgets.QWidget.resizeEvent(self, event)

    def question_preview_load_finished(self):
        #webView->page()->runJavaScript("document.documentElement.scrollWidth;",[=](QVariant result){
        #int newWidth=result.toInt();
        #webView->resize(newWidth,webView->height());
        #});

        #webView->page()->runJavaScript("document.documentElement.scrollHeight;",[=](QVariant result){
        #int newHeight=result.toInt();
        #webView->resize(webView->width(),newHeight);
        #});

        pass
        #self.required_question_size = \
        #    self.question_preview.page().currentFrame().contentsSize()
        #self.update_stretch_factors()

    def answer_preview_load_finished(self):
        pass
        #self.required_answer_size = \
        #    self.answer_preview.page().currentFrame().contentsSize()
        #self.update_stretch_factors()

    #
    # TMP workaround involving a heuristic to determine the height.
    #

    import re

    re_img = re.compile(r"""img src=\"file:///(.+?)\"(.*?)>""",
        re.DOTALL | re.IGNORECASE)

    def estimate_height(self, html):
        import math
        from mnemosyne.libmnemosyne.utils import expand_path, _abs_path
        from PIL import Image

        max_img_height = 0
        total_img_width = 0
        for match in self.re_img.finditer(html):
            img_file = match.group(1)
            if not _abs_path(img_file): # Linux issue
                img_file = "/" + img_file
            if not os.path.exists(img_file):
                print("Missing path", img_file)
                continue
            with Image.open(img_file) as im:
                width, height = im.size
                if height > max_img_height:
                    max_img_height = height
                total_img_width += width
        number_of_rows = math.ceil(total_img_width / self.question.width())
        return number_of_rows * max_img_height + 24

    def update_stretch_factors(self):
        if self.config()["QA_split"] != "adaptive":
            return
        if 0: # Using prerendered html
            # Correct the required heights of question and answer for the
            # presence of horizontal scrollbars.
            required_question_height = self.required_question_size.height()
            if self.required_question_size.width() > self.question.width():
                required_question_height += self.scrollbar_width
            required_answer_height = self.required_answer_size.height()
            if self.required_answer_size.width() > self.answer.width():
                required_answer_height += self.scrollbar_width
        else: # Tmp workaround using heuristic
            required_question_height = \
                self.estimate_height(self.question_text)
            required_answer_height = \
                self.estimate_height(self.answer_text)
        total_height_available = self.question.height() + self.answer.height()
        # If both question and answer fit in their own boxes, there is no need
        # to deviate from a 50/50 split.
        if required_question_height < total_height_available / 2 and \
            required_answer_height < total_height_available / 2:
            question_stretch = 50
            answer_stretch = 50
        # Don't be clairvoyant about the answer size, unless we will need
        # a non 50/50 split to start with.
        # If we are only showing the question, we try to limit 'surprising',
        # 'clairvoyant' stretches if they are not needed.
        elif not self.is_answer_showing:
            if required_question_height < total_height_available / 2:
                # No need to be clairvoyant.
                question_stretch = 50
                answer_stretch = 50
            else:
                # Make enough room for the question.
                question_stretch = required_question_height
                if required_question_height + required_answer_height \
                    <= total_height_available:
                    # Already have the stretch set-up to accomodate the answer,
                    # which makes the UI more relaxed (no need to have a
                    # different non 50/50 split once the answer is shown).
                    answer_stretch = required_answer_height
                else:
                    # But if we don't have enough space to show both the
                    # question and the answer, make sure the question gets
                    # all the space it can get now.
                    answer_stretch = 50
        # We are showing both question and answer.
        else:
            answer_stretch = required_answer_height
            if required_question_height + required_answer_height \
                    <= total_height_available:
                # If we have enough space, stretch in proportion to height.
                question_stretch = required_question_height
            else:
                # If we are previewing cards, go for a 50/50 split.
                if not self.used_for_reviewing:
                    question_stretch = 50
                    answer_stretch = 50
                # If we don't have enough space to show both the question and
                # the answer, try to give the answer all the space it needs.
                else:
                    answer_stretch = required_answer_height
                    question_stretch = total_height_available - answer_stretch
                    if question_stretch < 50:
                        question_stretch = 50
        self.setUpdatesEnabled(False)
        self.vertical_layout.setStretchFactor(\
            self.question_box, question_stretch + self.stretch_offset)
        self.vertical_layout.setStretchFactor(\
            self.answer_box, answer_stretch + self.stretch_offset)
        self.setUpdatesEnabled(True)

        # http://stackoverflow.com/questions/37527714/qt-qml-webview-resizes-really-slowly-when-window-resizing


    def silence_media(self, text):
        # Silence media, but make sure the player widget still shows to get
        # correct information about the geometry.
        text = text.replace("var soundFiles = new Array(",
                            "var soundFiles = new Array('off',")
        text = text.replace("<audio src=\"", "<audio src=\"off")
        text = text.replace("<video src=\"", "<video src=\"off")
        return text

    def set_question(self, text):
        #self.main_widget().show_information(text.replace("<", "&lt;"))
        self.question_text = text
        #self.question_preview.page().setPreferredContentsSize(\
        #    QtCore.QSize(self.question.size().width(), 1))
        #self.question_preview.setHtml(self.silence_media(text))
        #self.question_preview.show()

    def set_answer(self, text):
        #self.main_widget().show_information(text.replace("<", "&lt;"))
        self.answer_text = text

        #self.answer_preview.page().setPreferredContentsSize(\
        #    QtCore.QSize(self.answer.size().width(), 1))
        #self.answer_preview.setHtml(self.silence_media(text))
        #self.answer_preview.show()

    def reveal_question(self):
        self.question.setHtml(self.question_text, QtCore.QUrl("file://"))

    def reveal_answer(self):
        self.is_answer_showing = True
        self.update_stretch_factors()
        self.answer.setHtml(self.answer_text, QtCore.QUrl("file://"))
        # Forced repaint seems to make things snappier.
        self.question.repaint()
        self.answer.repaint()

    def clear_question(self):
        self.question.setHtml(self.empty())

    def clear_answer(self):
        self.is_answer_showing = False
        self.update_stretch_factors()
        self.answer.setHtml(self.empty())
        # Forced repaint seems to make things snappier.
        self.question.repaint()
        self.answer.repaint()


class ReviewWdgt(QtWidgets.QWidget, QAOptimalSplit, ReviewWidget, Ui_ReviewWdgt):

    auto_focus_grades = True
    number_keys_show_answer = True

    key_to_grade_map = {QtCore.Qt.Key_QuoteLeft: 0, QtCore.Qt.Key_0: 0,
            QtCore.Qt.Key_1: 1, QtCore.Qt.Key_2: 2, QtCore.Qt.Key_3: 3,
            QtCore.Qt.Key_4: 4, QtCore.Qt.Key_5: 5}

    def __init__(self, **kwds):
        super().__init__(**kwds)
        parent = self.main_widget()

        parent.setCentralWidget(self)
        self.setupUi(self)
        QAOptimalSplit.setup(self)

        # TODO: Move this to designer, once QButtonGroup elements can be set
        # to custom ID from designer.
        self.grade_buttons = QtWidgets.QButtonGroup()
        self.grade_buttons.addButton(self.grade_0_button, 0)
        self.grade_buttons.addButton(self.grade_1_button, 1)
        self.grade_buttons.addButton(self.grade_2_button, 2)
        self.grade_buttons.addButton(self.grade_3_button, 3)
        self.grade_buttons.addButton(self.grade_4_button, 4)
        self.grade_buttons.addButton(self.grade_5_button, 5)
        self.grade_buttons.buttonClicked[int].connect(self.grade_answer)
        # TODO: remove this once Qt bug is fixed.
        self.setTabOrder(self.grade_1_button, self.grade_2_button)
        self.setTabOrder(self.grade_2_button, self.grade_3_button)
        self.setTabOrder(self.grade_3_button, self.grade_4_button)
        self.setTabOrder(self.grade_4_button, self.grade_5_button)
        self.setTabOrder(self.grade_5_button, self.grade_0_button)
        self.setTabOrder(self.grade_0_button, self.grade_1_button)
        self.focus_widget = None
        self.sched = QtWidgets.QLabel("", parent.status_bar)
        self.notmem = QtWidgets.QLabel("", parent.status_bar)
        self.act = QtWidgets.QLabel("", parent.status_bar)
        parent.clear_status_bar()
        parent.add_to_status_bar(self.sched)
        parent.add_to_status_bar(self.notmem)
        parent.add_to_status_bar(self.act)
        parent.status_bar.setSizeGripEnabled(0)
        self.widget_with_last_selection = self.question
        self.question.selectionChanged.connect(self.selection_changed_in_q)
        self.answer.selectionChanged.connect(self.selection_changed_in_a)
        self.mplayer = QtCore.QProcess()
        self.media_queue = []
        self.setFocusPolicy(QtCore.Qt.StrongFocus)

        # When clicking out of the app, and clicking back on the web widgets,
        # the focus does not get properly restored, and for QWebEngineView, the
        # event handling for keypresses and focusIn events doesn't work, so
        # we do a crude workaround: https://bugreports.qt.io/browse/QTBUG-46251
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.restore_focus)
        self.timer.start(200)

    def deactivate(self):
        self.stop_media()
        ReviewWidget.deactivate(self)

    #def focusInEvent(self, event):
    #    self.restore_focus()
    #    super().focusInEvent(event)

    def changeEvent(self, event):
        if hasattr(self, "show_button"):
            show_button_text = self.show_button.text()
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslateUi(self)
        # retranslateUI resets the show button text to 'Show answer',
        # so we need to work around this.
        if hasattr(self, "show_button"):
            self.show_button.setText(_(show_button_text))
        QtWidgets.QWidget.changeEvent(self, event)

    def keyPressEvent(self, event):
        if event.key() in self.key_to_grade_map and not event.isAutoRepeat():
            # Use controller function rather than self.is_answer_showing to
            # deal with the map card type.
            if self.review_controller().is_question_showing():
                if self.number_keys_show_answer:
                    self.show_answer()
            else:
                self.grade_answer(self.key_to_grade_map[event.key()])
        elif event.key() == QtCore.Qt.Key_PageDown:
            self.scroll_down()
        elif event.key() == QtCore.Qt.Key_PageUp:
            self.scroll_up()
        elif event.key() == QtCore.Qt.Key_R and \
            event.modifiers() == QtCore.Qt.ControlModifier:
            self.review_controller().update_dialog(redraw_all=True) #Replay media.
        # QtWebengine issue.
        #elif event.key() == QtCore.Qt.Key_C and \
        #    event.modifiers() == QtCore.Qt.ControlModifier:
        #    self.copy()
        # Work around Qt issue.
        elif event.key() in [QtCore.Qt.Key_Backspace, QtCore.Qt.Key_Delete]:
            self.controller().delete_current_card()
        else:
            QtWidgets.QWidget.keyPressEvent(self, event)

    def empty(self):
        background = self.palette().color(QtGui.QPalette.Base).name()
        if self.review_controller().card:
            colour = self.config().card_type_property(\
            "background_colour", self.review_controller().card.card_type)
            if colour:
                background = ("%X" % colour)[2:] # Strip alpha.
        return """
        <html><head>
        <style type="text/css">
        table { height: 100%; }
        body  { background-color: """ + background + """;
                margin: 0;
                padding: 0;
                border: thin solid #8F8F8F; }
        </style></head>
        <body><table><tr><td>
        </td></tr></table></body></html>"""

    def scroll_down(self):
        return
        # TODO: reimplement after webkit is back.

        if self.review_controller().is_question_showing() or \
           self.review_controller().card.fact_view.a_on_top_of_q:
            frame = self.question.page().mainFrame()
        else:
            frame = self.answer.page().mainFrame()
        x, y = frame.scrollPosition().x(), frame.scrollPosition().y()
        y += int(0.9*(frame.geometry().height()))
        #frame.scroll(x, y) # Seems buggy 20111121.
        frame.evaluateJavaScript("window.scrollTo(%d, %d);" % (x, y))

    def scroll_up(self):
        return
        # TODO: reimplement after webkit is back.

        if self.review_controller().is_question_showing() or \
           self.review_controller().card.fact_view.a_on_top_of_q:
            frame = self.question.page().mainFrame()
        else:
            frame = self.answer.page().mainFrame()
        x, y = frame.scrollPosition().x(), frame.scrollPosition().y()
        y -= int(0.9*(frame.geometry().height()))
        #frame.scroll(x, y)  # Seems buggy 20111121.
        frame.evaluateJavaScript("window.scrollTo(%d, %d);" % (x, y))

    def selection_changed_in_q(self):
        self.widget_with_last_selection = self.question

    def selection_changed_in_a(self):
        self.widget_with_last_selection = self.answer

    def copy(self):
        self.widget_with_last_selection.pageAction(\
            QtWebEngineKit.QWebEnginePage.Copy).trigger()

    def show_answer(self):
        self.review_controller().show_answer()

    def grade_answer(self, grade):
        self.review_controller().grade_answer(grade)

    def set_question_box_visible(self, is_visible):
        if is_visible:
            self.question.show()
            self.question_label.show()
        else:
            self.question.hide()
            self.question_label.hide()

    def set_answer_box_visible(self, is_visible):
        if is_visible:
            self.answer.show()
            self.answer_label.show()
        else:
            self.answer.hide()
            self.answer_label.hide()

    def set_question_label(self, text):
        self.question_label.setText(text)

    def restore_focus(self):
        # After clicking on the question or the answer, that widget grabs the
        # focus, so that the keyboard shortcuts no longer work. This functions
        # is used to set the focus back to the correct widget.
        if self.question.hasFocus() or self.answer.hasFocus():
            if self.focus_widget:
                self.focus_widget.setDefault(True)
                self.focus_widget.setFocus()

    def update_show_button(self, text, is_default, is_enabled):
        self.show_button.setText(text)
        self.show_button.setEnabled(is_enabled)
        if is_default:
            self.show_button.setDefault(True)
            self.show_button.setFocus()
            self.focus_widget = self.show_button

    def set_grades_enabled(self, is_enabled):
        self.grades.setEnabled(is_enabled)

    def set_grade_enabled(self, grade, is_enabled):
        self.grade_buttons.button(grade).setEnabled(is_enabled)

    def set_default_grade(self, grade):
        if self.auto_focus_grades:
            # On Windows, we seem to need to clear the previous default
            # first.
            for grade_i in range(6):
                self.grade_buttons.button(grade_i).setDefault(False)
            self.grade_buttons.button(grade).setDefault(True)
            self.grade_buttons.button(grade).setFocus()
            self.focus_widget = self.grade_buttons.button(grade)

    def set_grades_title(self, text):
        self.grades.setTitle(text)

    def set_grade_text(self, grade, text):
        self.grade_buttons.button(grade).setText(text)

    def set_grade_tooltip(self, grade, text):
        self.grade_buttons.button(grade).setToolTip(text)

    def update_status_bar_counters(self):
        scheduled_count, non_memorised_count, active_count = \
            self.review_controller().counters()
        self.sched.setText(_("Scheduled: %d ") % scheduled_count)
        self.notmem.setText(_("Not memorised: %d ") % non_memorised_count)
        self.act.setText(_("Active: %d ") % active_count)

    def play_media(self, filename, start=None, stop=None):
        if start is None:
            start = 0
        if stop is None:
            stop = 999999
        self.media_queue.append((filename, start, stop))
        if self.mplayer.state() != QtCore.QProcess.Running:
            self.play_next_file()

    def play_next_file(self):
        filename, start, stop = self.media_queue.pop(0)
        duration = stop - start
        if duration > 400:
            duration -= 300 # Compensate for mplayer overshoot.
        self.mplayer = QtCore.QProcess()
        self.mplayer.finished.connect(self.done_playing)
        if sys.platform == "win32":
            command = "mplayer.exe -slave -ao win32 -quiet \"" + filename + \
                "\" -ss " + str(start) + " -endpos " + str(duration)
        elif sys.platform == "darwin":
            # e.g. /path/to/Mnemosyne.app/Contents/MacOS/mnemosyne/pyqt_ui/review_wdgt.py
            SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))
            # e.g. /path/to/Mnemosyne.app/Contents
            CONTENTS_FOLDER = SCRIPT_PATH[:SCRIPT_PATH.index("/MacOS")]
            # e.g. /path/to/Mnemosyne.app/Contents/MacOS/mplayer
            MPLAYER_PATH = CONTENTS_FOLDER + "/MacOS/mplayer"

            command = "{} -slave -vo null -ao coreaudio -quiet \"{}\" -ss {} -endpos {}".format(
                MPLAYER_PATH, filename, str(start), str(duration)
            )
        else:
            command = "mplayer -slave -quiet \"" + filename + \
                "\" -ss " + str(start) + " -endpos " + str(duration)
        self.mplayer.start(command)

    def done_playing(self, result):
        if len(self.media_queue) >= 1:
            self.play_next_file()

    def stop_media(self):
        if self.mplayer and self.mplayer.state() == QtCore.QProcess.Running:
            self.mplayer.write(b"quit\n");
        self.media_queue = []

    def redraw_now(self):
        self.repaint()
        self.parent().repaint()



#
# default_criterion.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.criterion import Criterion


class DefaultCriterion(Criterion):

    criterion_type = "default"

    def __init__(self, component_manager, id=None):
        Criterion.__init__(self, component_manager, id)
        # (card_type.id, fact_view.id):
        self.deactivated_card_type_fact_view_ids = set()
        # We work with _ids instead of ids for speed.
        self._tag_ids_active = set()
        self._tag_ids_forbidden = set()

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        return self.deactivated_card_type_fact_view_ids == \
                other.deactivated_card_type_fact_view_ids \
            and self._tag_ids_active == other._tag_ids_active \
            and self._tag_ids_forbidden == other._tag_ids_forbidden

    def is_empty(self):
        # Check card types.
        counter = 0
        for card_type in self.card_types():
            counter += len(card_type.fact_views)
        if len(self.deactivated_card_type_fact_view_ids) == counter:
            return True
        # Active tags.
        elif len(self._tag_ids_active) == 0:
            return True
        # Forbidden tags.
        elif len(self.database().tags()) == len(self._tag_ids_forbidden):
            return True
        else:
            return False

    def apply_to_card(self, card):
        card.active = False
        for tag in card.tags:
            if tag._id in self._tag_ids_active:
                card.active = True
                break
        if (card.card_type.id, card.fact_view.id) in \
           self.deactivated_card_type_fact_view_ids:
            card.active = False
        for tag in card.tags:
            if tag._id in self._tag_ids_forbidden:
                card.active = False
                break

    def active_tag_added(self, tag):
        self._tag_ids_active.add(tag._id)

    def deactivated_tag_added(self, tag):
        # Since the UI only supports criteria which either have active tags or
        # forbidden tags, but not both, we need to maintain that situation when
        # adding tags.
        if len(self._tag_ids_forbidden) != 0:
            self._tag_ids_forbidden.add(tag._id)
        else:
            pass

    def is_tag_active(self, tag):
        return (tag._id in self._tag_ids_active)
    
    def tag_deleted(self, tag):
        self._tag_ids_active.discard(tag._id)
        self._tag_ids_forbidden.discard(tag._id)

    def active_card_type_added(self, card_type):
        pass

    def deactivated_card_type_added(self, card_type):
        for fact_view in card_type.fact_views:
            self.deactivated_card_type_fact_view_ids.add(\
                (card_type.id, fact_view.id))

    def card_type_deleted(self, card_type):
        for fact_view in card_type.fact_views:
            self.deactivated_card_type_fact_view_ids.discard(\
                (card_type.id, fact_view.id))

    def data_to_string(self):
        return repr((self.deactivated_card_type_fact_view_ids,
                     self._tag_ids_active,
                     self._tag_ids_forbidden))

    def set_data_from_string(self, data_string):
        data = eval(data_string)
        self.deactivated_card_type_fact_view_ids = data[0]
        self._tag_ids_active = data[1]
        self._tag_ids_forbidden = data[2]

    # To send the criteria across, we need to convert from _ids ids first.

    def data_to_sync_string(self):
        active_tag_ids = set()
        for _tag_id in self._tag_ids_active:
            tag = self.database().tag(_tag_id, is_id_internal=True)
            active_tag_ids.add(tag.id)
        forbidden_tag_ids = set()
        for _tag_id in self._tag_ids_forbidden:
            tag = self.database().tag(_tag_id, is_id_internal=True)
            forbidden_tag_ids.add(tag.id)
        return repr((self.deactivated_card_type_fact_view_ids,
                     active_tag_ids, forbidden_tag_ids))

    def set_data_from_sync_string(self, data_string):
        data = eval(data_string)
        self.deactivated_card_type_fact_view_ids = data[0]
        active_tag_ids = data[1]
        forbidden_tag_ids = data[2]
        self._tag_ids_active = set()
        for tag_id in active_tag_ids:
            tag = self.database().tag(tag_id, is_id_internal=False)
            self._tag_ids_active.add(tag._id)
        self._tag_ids_forbidden = set()
        for tag_id in forbidden_tag_ids:
            tag = self.database().tag(tag_id, is_id_internal=False)
            self._tag_ids_forbidden.add(tag._id)



#
# test_tsv_import.py <Peter.Bienstman@UGent.be>
#

import os
import shutil
from nose.tools import raises

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne import Mnemosyne
from openSM2sync.log_entry import EventTypes
from mnemosyne.libmnemosyne.ui_components.dialogs import ImportDialog
from mnemosyne.libmnemosyne.ui_components.main_widget import MainWidget

last_error = ""

class Widget(MainWidget):

    def show_error(self, message):
        global last_error
        print(message)
        last_error = message

class TestDBImport(MnemosyneTest):

    def setup(self):
        self.initialise_data_dir()
        self.mnemosyne = Mnemosyne(upload_science_logs=False, interested_in_old_reps=True,
                    asynchronous_database=True)
        self.mnemosyne.components.insert(0,
           ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator", "GetTextGuiTranslator"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = \
            [("mnemosyne_test", "TestReviewWidget")]
        self.mnemosyne.components.append(\
            ("test_db_import", "Widget"))
        self.mnemosyne.initialise(os.path.abspath("dot_test"), automatic_upgrades=False)
        self.review_controller().reset()
        self.merge_db_path = os.path.join(os.getcwd(), "tests", "files", "basedir_to_merge", "to_merge.db")
        self.merge_db_tmppath = os.path.join(os.path.dirname(self.merge_db_path), "to_merge_tmp.db")
        shutil.copy2(self.merge_db_path, self.merge_db_tmppath)

    def db_importer(self):
        for format in self.mnemosyne.component_manager.all("file_format"):
            if format.__class__.__name__ == "Mnemosyne2Db":
                return format

    def test_1(self):
        print(os.getcwd())
        filename = os.path.join(os.getcwd(), "tests", "files", "basedir_to_merge", "config.db")
        self.db_importer().do_import(filename)
        assert "configuration database" in last_error

    def test_2(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        old_card = self.controller().create_new_cards(fact_data, card_type,
                                 grade=-1, tag_names=["default"])[0]
        assert len([self.database().cards()]) == 1

        global last_error
        last_error = ""
        self.db_importer().do_import(self.merge_db_tmppath)
        assert last_error == ""
        db = self.database()
        assert db.con.execute("select count() from log where event_type != 26").fetchone()[0] == 258
        self.review_controller().reset()
        assert self.database().card_count() == 7
        assert self.database().active_count() == 6
        assert self.database().fact_count() == 5
        card_type = self.database().card_type("2::new clone", is_id_internal=False)
        assert self.config().card_type_property("background_colour", card_type) == 4278233600

    def teardown(self):
        MnemosyneTest.teardown(self)
        if os.path.exists(self.merge_db_path + "-journal"):
            os.remove(self.merge_db_path + "-journal")
        if os.path.exists(self.merge_db_tmppath):
            os.remove(self.merge_db_tmppath)
        if os.path.exists(self.merge_db_tmppath + "-journal"):
            os.remove(self.merge_db_tmppath + "-journal")



#
# test_tag_tree.py <Peter.Bienstman@UGent.be>
#

from mnemosyne_test import MnemosyneTest
from mnemosyne.libmnemosyne.criteria.default_criterion import DefaultCriterion


class TestTagTree(MnemosyneTest):

    def test_1(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=[])[0]
        self.controller().save_file()
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        assert len(list(self.tree.keys())) == 2
        assert self.tree['__ALL__'] == ['__UNTAGGED__']

    def test_2(self):
        fact_data = {"f": "question",
                     "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["tag_1"])[0]
        self.controller().save_file()
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        assert len(list(self.tree.keys())) == 3
        assert self.tree['__ALL__'] == ['tag_1', '__UNTAGGED__']

    def test_3(self):
        fact_data = {"f": "question", "b": "answer"}
        card_type = self.card_type_with_id("1")
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a"])[0]
        fact_data = {"f": "question2", "b": "answer2"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["Z"])[0]
        fact_data = {"f": "question3",  "b": "answer3"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::b"])[0]
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::c"])[0]
        fact_data = {"f": "question5",  "b": "answer5"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["b::c::d"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert self.tree.card_count_for_node["b"] == 1
        assert self.tree.card_count_for_node["b::c"] == 1
        assert self.tree.card_count_for_node["b::c::d"] == 1
        assert self.tree.nodes() == \
               ["a", "a::Untagged", "a::b", "a::c", "b", "b::c", "b::c::d",
                "Z", "__UNTAGGED__"]

    def test_4(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::b"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a"])[0]
        fact_data = {"f": "question2", "b": "answer2"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["Z"])[0]
        fact_data = {"f": "question3",  "b": "answer3"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::c"])[0]
        fact_data = {"f": "question5",  "b": "answer5"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["b::c::d"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert self.tree.card_count_for_node["b"] == 1
        assert self.tree.card_count_for_node["b::c"] == 1
        assert self.tree.card_count_for_node["b::c::d"] == 1

        self.tree.rename_node("Z", "Z::Z")
        self.tree.rename_node("b::c", "b::cc")
        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["Z::Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert self.tree.card_count_for_node["b"] == 1
        assert self.tree.card_count_for_node["b::cc"] == 1
        assert self.tree.card_count_for_node["b::cc::d"] == 1

    def test_5(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::b"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a"])[0]
        fact_data = {"f": "question2", "b": "answer2"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["Z"])[0]
        fact_data = {"f": "question3",  "b": "answer3"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::c"])[0]
        fact_data = {"f": "question5",  "b": "answer5"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["b::c::d"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert self.tree.card_count_for_node["b"] == 1
        assert self.tree.card_count_for_node["b::c"] == 1
        assert self.tree.card_count_for_node["b::c::d"] == 1

        self.tree.rename_node("b::c", "b")
        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert self.tree.card_count_for_node["b"] == 1
        assert self.tree.card_count_for_node["b::d"] == 1

    def test_rename_to_existing_tag(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["tag1"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["tag2"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        assert len(self.database().tags()) == 3
        self.tree.rename_node("tag1", "tag2")
        assert self.tree.card_count_for_node["tag2"] == 2
        assert len(self.database().tags()) == 2

    def test_rename_to_existing_tag_2(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["Xx::vb::test", "Xx::aa::vb::test"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        assert len(self.database().tags()) == 3
        self.tree.rename_node("Xx::aa::vb::test", "Xx::vb::test")
        assert self.tree.card_count_for_node["Xx::vb::test"] == 1
        assert "Xx::aa::vb::test" not in self.tree.card_count_for_node
        assert "," not in  self.database().con.execute(\
            "select tags from cards where _id=?", (card._id,)).fetchone()[0]
        assert self.database().con.execute(\
            "select count() from tags_for_card").fetchone()[0] == 1
        assert len(self.database().tags()) == 2

    def test_rename_to_empty(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["tag1"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["tag2"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        self.tree.rename_node("tag1", "")
        assert self.tree.card_count_for_node["__UNTAGGED__"] == 1
        assert len(self.database().tags()) == 2

    def test_rename_to_empty_2(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["A::tag1"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        self.tree.rename_node("A", "")
        assert self.tree.card_count_for_node["tag1"] == 1
        assert len(self.database().tags()) == 2

    def test_rename_to_empty_3(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["A::tag1"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["A"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        self.tree.rename_node("A", "")
        assert self.tree.card_count_for_node["__UNTAGGED__"] == 1
        assert self.tree.card_count_for_node["tag1"] == 1
        assert len(self.database().tags()) == 2

    def test_rename_to_forbidden(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["A::tag1"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["A"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        self.tree.rename_node("A", "__UNTAGGED__")
        assert "__UNTAGGED__" in self.tree.card_count_for_node

    def test_delete(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::b"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a"])[0]
        fact_data = {"f": "question2", "b": "answer2"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["Z"])[0]
        fact_data = {"f": "question3",  "b": "answer3"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::c"])[0]
        fact_data = {"f": "question5",  "b": "answer5"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["b::c::d"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert self.tree.card_count_for_node["b"] == 1
        assert self.tree.card_count_for_node["b::c"] == 1
        assert self.tree.card_count_for_node["b::c::d"] == 1

        self.tree.delete_subtree("b::c")

        card = self.database().card(card._id, is_id_internal=True)
        assert card.tag_string() == ""
        self.database().con.execute("select tags from cards where _id=?",
            (card._id, )).fetchone()[0] == ""

        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert self.tree.card_count_for_node["__UNTAGGED__"] == 1
        assert "b" not in self.tree.card_count_for_node
        assert "b::c" not in self.tree.card_count_for_node
        assert "b::c::d" not in self.tree.card_count_for_node

    def test_delete_2(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::b"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a"])[0]
        fact_data = {"f": "question2", "b": "answer2"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["Z"])[0]
        fact_data = {"f": "question3",  "b": "answer3"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["a::c"])[0]
        fact_data = {"f": "question5",  "b": "answer5"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["b::c::d", "b"])[0]
        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)

        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert self.tree.card_count_for_node["b"] == 1
        assert self.tree.card_count_for_node["b::c"] == 1
        assert self.tree.card_count_for_node["b::c::d"] == 1

        self.tree.delete_subtree("b::c")
        card = self.database().card(card._id, is_id_internal=True)
        assert card.tag_string() == ""
        self.database().con.execute("select tags from cards where _id=?",
            (card._id, )).fetchone()[0] == "b"

        assert self.tree.card_count_for_node["__ALL__"] == 5
        assert self.tree.card_count_for_node["a"] == 3
        assert self.tree.card_count_for_node["Z"] == 1
        assert self.tree.card_count_for_node["a::b"] == 1
        assert self.tree.card_count_for_node["a::c"] == 1
        assert "b" not in self.tree.card_count_for_node
        assert "__UNTAGGED__" in self.tree.card_count_for_node
        assert "b::c" not in self.tree.card_count_for_node
        assert "b::c::d" not in self.tree.card_count_for_node

    def test_count_1(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["b"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["b", "b"])[0]

        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)

        assert self.tree.card_count_for_node["__ALL__"] == 2

    def test_count_2(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question4",  "b": "answer4"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["X::a"])[0]
        fact_data = {"f": "question", "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["X::a", "X::b"])[0]

        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)

        assert self.tree.card_count_for_node["__ALL__"] == 2
        assert self.tree.card_count_for_node["X"] == 2

    def test_delete_forbidden(self):
        card_type = self.card_type_with_id("1")
        fact_data = {"f": "question",  "b": "answer"}
        card = self.controller().create_new_cards(fact_data, card_type,
            grade=-1, tag_names=["forbidden"])[0]
        assert self.database().active_count() == 1

        c = DefaultCriterion(self.mnemosyne.component_manager)
        c.deactivated_card_type_fact_view_ids = set()
        c._tag_ids_active = set([self.database().get_or_create_tag_with_name("active")._id, 1])
        c._tag_ids_forbidden = set([self.database().get_or_create_tag_with_name("forbidden")._id])
        self.database().set_current_criterion(c)
        assert self.database().active_count() == 0

        from mnemosyne.libmnemosyne.tag_tree import TagTree
        self.tree = TagTree(self.mnemosyne.component_manager)
        self.tree.delete_subtree("forbidden")
        assert self.database().active_count() == 1



# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
# Based off Kieran Clancy's initial implementation.

import re
from ..hooks import addHook

r = r' ?([^ >]+?)\[(.+?)\]'
ruby = r'<ruby><rb>\1</rb><rt>\2</rt></ruby>'

def noSound(repl):
    def func(match):
        if match.group(2).startswith("sound:"):
            # return without modification
            return match.group(0)
        else:
            return re.sub(r, repl, match.group(0))
    return func

def _munge(s):
    return s.replace("&nbsp;", " ")

def kanji(txt, *args):
    return re.sub(r, noSound(r'\1'), _munge(txt))

def kana(txt, *args):
    return re.sub(r, noSound(r'\2'), _munge(txt))

def furigana(txt, *args):
    return re.sub(r, noSound(ruby), _munge(txt))

def install():
    addHook('fmod_kanji', kanji)
    addHook('fmod_kana', kana)
    addHook('fmod_furigana', furigana)



#
# review_widget.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.ui_component import UiComponent


class ReviewWidget(UiComponent):
    
    """Describes the interface that the review widget needs to implement
    in order to be used by the review controller.

    Note that also the review widget is instantiated late, even though we
    always need one right from the start of the program. The reason is that
    there could be many review widgets defined through plugins, and
    instantiating them all at the start of the program could be slow,
    especially on mobile devices.

    'set_question' and 'set_answer' do not yet show the question and the
    answer. This allows the review widget to tweak the layout based on the
    answer even before it is shown.

    """

    component_type = "review_widget"
        
    def set_question_box_visible(self, is_visible):
        pass
        
    def set_answer_box_visible(self, is_visible):
        pass
        
    def set_question_label(self, text):
        pass
        
    def set_question(self, text):
        pass
        
    def set_answer(self, text):
        pass
    
    def reveal_question(self):
        pass
        
    def reveal_answer(self):
        pass
    
    def clear_question(self): 
        pass
        
    def clear_answer(self): 
        pass
        
    def update_show_button(self, text, is_default, is_enabled): 
        pass

    def set_grades_enabled(self, is_enabled):

        """Enable whole grade area, including title."""
        
        pass
    
    def set_grade_enabled(self, grade, is_enabled):

        """Enable just a single grade button."""
        
        pass
    
    def set_default_grade(self, grade):
        pass
        
    def set_grades_title(self, text): 
        pass
            
    def set_grade_text(self, grade, text): 
        pass
            
    def set_grade_tooltip(self, grade, text): 
        pass

    def update_status_bar_counters(self):
        pass
    
    def play_media(self, filename, start=None, stop=None):
        pass    
    
    def stop_media(self):
        pass

    def redraw_now(self):

        """Tell the widget that now would be a good time to redraw
        immediately. Some toolkits queue redraw events until there is no more
        processing going on, but this defeats libmnemosyne's optimisation of
        trying to show the next question before the grading of the previous
        question is completed.

        We don't provide a default 'pass' implementation, to force front-end
        writers to think about this.

        """

        raise NotImplementedError



#
# main_wdgt.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.UDP_server.UDP_widget import UDP_Widget
from mnemosyne.libmnemosyne.ui_components.main_widget import MainWidget


class MainWdgt(MainWidget, UDP_Widget):

    def __init__(self, component_manager):
        MainWidget.__init__(self, component_manager)

    def set_window_title(self, text):
        self.callback(text)

    def show_information(self, text):
        self.callback(text)

    def show_question(self, text, option0, option1, option2):
        self.callback(text, option0, option1, option2)
        return int(self.read_from_socket())

    def show_error(self, text):
        self.callback(text)

    def get_filename_to_open(self, path, filter, caption=""):
        self.callback(path, filter, caption)
        return self.read_from_socket()

    def get_filename_to_save(self, path, filter, caption=""):
        self.callback(path, filter, caption)
        return self.read_from_socket()

    def set_status_bar_message(self, text):
        self.callback(text)

    def set_progress_text(self, text):
        self.callback(text)

    def set_progress_range(self, maximum):
        self.callback(maximum)

    def set_progress_update_interval(self, update_interval):
        self.callback(update_interval)

    def increase_progress(self, value):
        self.callback(value)

    def set_progress_value(self, value):
        self.callback(value)

    def close_progress(self):
        self.callback()

    def enable_edit_current_card(self, is_enabled):
        self.callback(is_enabled)

    def enable_delete_current_card(self, is_enabled):
        self.callback(is_enabled)

    def enable_browse_cards(self, is_enabled):
        self.callback(is_enabled)




#
# after_repetition.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.hook import Hook
from mnemosyne.libmnemosyne.plugin import Plugin


class Grade5DetectionHook(Hook):

    used_for = "after_repetition"

    def run(self, card):
        if card.grade == 5:
            self.main_widget().show_information("This was an easy card!")


class AfterRepetitionPlugin(Plugin):
    
    name = "Grade 5 detection"
    description = "Notice when a card is given grade 5."   
    components = [Grade5DetectionHook]
    supported_API_level = 3
    

# Register plugin.

from mnemosyne.libmnemosyne.plugin import register_user_plugin
register_user_plugin(AfterRepetitionPlugin)




#
# google_pronouncer.py <Peter.Bienstman@UGent.be>
#

from gtts import gTTS

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.utils import expand_path
from mnemosyne.libmnemosyne.pronouncer import Pronouncer

class GooglePronouncer(Pronouncer):

    used_for = "bn", "bn-BD", "bn-ID", "zh", "zh-CN", "zh-TW", "cs", "da", \
        "nl", "en", "en-GB", "en-US", "en-IN", "en-AU", "et", "fi", "fr", \
        "fr-FR", "fr-CA", "de", "el", "hi", "hu", "id", "it", "ja", "jw", \
        "km", "ko", "ne", "no", "pl", "pt", "pt-PT", "pt-BR", "ro", "ru", \
        "si", "sk", "es", "es-ES", "es-US", "su", "sv", "th", "tr", "uk", \
        "vi", "ar", "is"

    popup_menu_text = "Insert Google text-to-speech..."

    # https://en.wikipedia.org/wiki/Google_Text-to-Speech

    def download_tmp_audio_file(self, card_type, foreign_text):

        """Returns a temporary filename with the audio."""

        language_id = self.config().card_type_property(\
            "sublanguage_id", card_type)
        if not language_id:
            language_id = self.config().card_type_property(\
                "language_id", card_type)
        tts = gTTS(foreign_text, lang=language_id)
        filename = expand_path("__GTTS__TMP__.mp3", self.database().media_dir())
        tts.save(filename)
        return filename



#
# SM2_controller_cramming.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.review_controllers.SM2_controller \
     import SM2Controller
from mnemosyne.libmnemosyne.schedulers.cramming import Cramming


class SM2ControllerCramming(SM2Controller):

    def grade_answer(self, grade):
        self.flush_sync_server()
        card_to_grade = self.card
        old_grade = card_to_grade.grade
        self.update_counters(old_grade, grade)
        self.rep_count += 1
        if self.scheduler().is_prefetch_allowed(card_to_grade):
            self.show_new_question()
            interval = self.scheduler().grade_answer(card_to_grade, grade)
            self.database().update_card(card_to_grade, repetition_only=True)
            if self.rep_count % self.config()["save_after_n_reps"] == 0:
                self.database().save()
        else:
            interval = self.scheduler().grade_answer(card_to_grade, grade)
            self.database().update_card(card_to_grade, repetition_only=True)
            if self.rep_count % self.config()["save_after_n_reps"] == 0:
                self.database().save()
            self.show_new_question()
        self.review_widget().update_status_bar_counters()
        if self.config()["show_intervals"] == "status_bar":
            self.main_widget().update_status_bar_message(_("Returns in") + \
                " " + str(interval) + _(" day(s)."))

    def counters(self):
        db = self.database()
        max_ret_reps = 1 if self.new_only else -1
        return db.scheduler_data_count(Cramming.WRONG, max_ret_reps), \
            db.scheduler_data_count(Cramming.UNSEEN, max_ret_reps),  \
            db.active_count()

    def reload_counters(self):
        pass

    def update_counters(self, old_grade, new_grade):
        pass

    def update_grades_area(self):
        self.review_widget().set_grades_enabled(self.grades_enabled)
        if self.grades_enabled:
            self.review_widget().set_default_grade(5)



#
# M_sided.py <Peter.Bienstman@UGent.be>
#

import os
import sys
import copy

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.plugin import Plugin
from mnemosyne.libmnemosyne.card_type import CardType
from mnemosyne.libmnemosyne.filters.escape_to_html import EscapeToHtml


class MSided(CardType):

    """Abstract parent class for M-sided card types with Anki syntax."""

    id = "7"
    name = _("M-sided")
    hidden_from_UI = True

    # Need to be overwritten with correct instance variables at creation.
    fact_keys_and_names = []
    fact_views = []
    unique_fact_keys = []
    required_fact_keys = []

    def __init__(self, component_manager):
        super().__init__(component_manager)
        # Add Anki files to Python path.
        path = os.path.join(os.path.dirname(sys.modules['mnemosyne'].__file__),
                            "libmnemosyne", "renderers")
        if path not in sys.path:
            sys.path.append(path)
        from mnemosyne.libmnemosyne.renderers.anki_renderer import AnkiRenderer
        self.renderer = AnkiRenderer(component_manager)

    def render_question(self, card, render_chain="default", **render_args):
        return self._render(card, render_chain, render_QA="Q", **render_args)

    def render_answer(self, card, render_chain="default", **render_args):
        # If we cannot isolate the question from the answer, and we need
        # the 'FrontSide' info, calculate this info first.
        if self.config()["QA_split"] == "single_window" or \
                "<hr id=answer>" not in card.fact_view.extra_data["afmt"]:
            if "FrontSide" in card.fact_view.extra_data["afmt"]:
                new_render_args = copy.copy(render_args)
                new_render_args["body_only"] = True
                render_args["FrontSide"] = \
                    self.render_question(card, render_chain, **new_render_args)
        return self._render(card, render_chain, render_QA="A", **render_args)

    def _render(self, card, render_chain="default", **render_args):
        # Because there is no list of fact keys for q or a for this card type,
        # we need to run the filters after assembling the content, not before.
        html = self.renderer.render(\
            card, card.fact.data, render_chain, **render_args)
        if "body_only" in render_args and render_args["body_only"] == True:
            return html  # Filters will be run later.
        for filter in self.render_chain(render_chain)._filters:
            # EscapeToHtml introduces <br> inside css.
            if filter.__class__ != EscapeToHtml:
                html = filter.run(html, card, fact_key=None, **render_args)
        return html




#
# anki_renderer.py <Peter.Bienstman@UGent.be>
#

import re
import copy

from mnemosyne.libmnemosyne.renderer import Renderer
from mnemosyne.libmnemosyne.card_types.M_sided import MSided
from mnemosyne.libmnemosyne.renderers.anki.template import render as _render


class AnkiRenderer(Renderer):

    """Renders the question or the answer using the Anki syntax."""

    used_for = MSided

    def render(self, card, filtered_fact_data, render_chain, **render_args):
        card_type = card.card_type
        extra_data = card.fact_view.extra_data
        extra_data["ord"] = card.extra_data.get("ord", 0)
        fields = {}
        for fact_key, fact_key_name in card_type.fact_keys_and_names:
            fields[fact_key_name] = filtered_fact_data.get(fact_key, "")
        fields["Tags"] = "" # Mnemosyne shows tags elsewhere.
        fields["Type"] = card_type.name
        fields["Deck"] = ""
        fields["Subdeck"] = ""
        fields["Card"] = card.fact_view.name
        fields["c%d" % (extra_data["ord"] + 1)] = "1"
        # Determine template.
        if render_args["render_QA"] == "Q":
            if render_chain in ["plain_text", "card_browser"] \
               and extra_data["bqfmt"]:
                template = extra_data["bqfmt"]
            else:
                template = extra_data["qfmt"]
            template = re.sub("{{(?!type:)(.*?)cloze:", r"{{\1cq-%d:" \
                                % (extra_data["ord"] + 1), template)
            template = template.replace("<%cloze:", "<%%cq:%d:" % (
                    extra_data["ord"] + 1))
        else:
            if render_chain in ["plain_text", "card_browser"] and \
               extra_data["bafmt"]:
                    template = extra_data["bafmt"]
            else:
                template = extra_data["afmt"]
            # If possible, strip the question part from the template, so that
            # we can display Q and A in a separate window.
            if self.config()["QA_split"] != "single_window" or \
               render_chain in ["plain_text", "card_browser"]:
                if "<hr id=answer>" in template:
                    template = template.split("<hr id=answer>", 1)[1].strip()
            # Deal with clozes.
            template = re.sub("{{(.*?)cloze:", r"{{\1ca-%d:" \
                              % (extra_data["ord"]+1), template)
            template = template.replace("<%cloze:", "<%%ca:%d:" % (
                    extra_data["ord"] + 1))
            if "FrontSide" in render_args:
                fields["FrontSide"] = render_args["FrontSide"].\
                    replace("audio src", "audio_off src")
        # Determine colours and fonts.
        for fact_key, fact_key_name in card.card_type.fact_keys_and_names:
            style = ""
            colour = self.config().card_type_property(\
                "font_colour", card_type, fact_key)
            if colour:
                colour_string = ("%X" % colour)[2:] # Strip alpha.
                style += "color: #%s; " % colour_string
            font_string = self.config().card_type_property(\
                "font", card.card_type, fact_key)
            if font_string:
                if font_string.count(",") == 10:
                    family,size,x,x,w,i,u,s,x,x,x = font_string.split(",")
                else:
                    family,size,x,x,w,i,u,s,x,x = font_string.split(",")
                style += "font-family: '%s'; " % family
                style += "font-size: %spt; " % size
                if w == "25":
                    style += "font-weight: light; "
                if w == "75":
                    style += "font-weight: bold; "
                if i == "1":
                    style += "font-style: italic; "
                if i == "2":
                    style += "font-style: oblique; "
                if u == "1":
                    style += "text-decoration: underline; "
                if s == "1":
                    style += "text-decoration: line-through; "
                fields[fact_key_name] = "<span style=\"%s\">%s</span>" % \
                    (style, fields[fact_key_name])
        # Hide {{type:...}} fields.
        template = re.sub("{{type:.+}}", "", template)
        # Hide {{hint:...}} fields in card browser.
        if render_chain in ["plain_text", "card_browser"]:
            template = re.sub("{{hint:.+}}", "", template)
        body = _render(template, fields)
        # Some heuristic to have a decent display in the card browser.
        if render_chain in ["plain_text", "card_browser"] or \
           ("body_only" in render_args and render_args["body_only"] == True):
            if "bqfmt" not in extra_data or not extra_data["bqfmt"]:
                body = body.replace("font-size", "font-size-off")
                body = body.replace("\n", "").replace("<br", "<br_off")
            return body
        else:
            return """
            <!doctype html>
            <html>
            <head><style>%s</style></head>
            <body class="card">%s</body>
            </html>""" % (card_type.extra_data["css"], body)




#
# criterion_wdgt_default.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.ui_components.criterion_widget \
     import CriterionWidget
from mnemosyne.libmnemosyne.criteria.default_criterion import DefaultCriterion
from mnemosyne.pyqt_ui.tag_tree_wdgt import TagsTreeWdgt
from mnemosyne.pyqt_ui.card_type_tree_wdgt import CardTypesTreeWdgt
from mnemosyne.pyqt_ui.ui_criterion_wdgt_default import Ui_DefaultCriterionWdgt


class DefaultCriterionWdgt(QtWidgets.QWidget, CriterionWidget,
                           Ui_DefaultCriterionWdgt):

    """Note that this dialog can support active tags and forbidden tags,
    but not at the same time, in order to keep the interface compact.

    """

    used_for = DefaultCriterion

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.parent = kwds["parent"]
        component_manager = kwds["component_manager"]
        self.setupUi(self)
        self.card_type_tree_wdgt = CardTypesTreeWdgt(\
            component_manager=component_manager, parent=self)
        # Bug in Qt: need to explicitly reset the text of this label.
        self.label.setText(_("Activate cards from these card types:"))
        self.gridLayout.addWidget(self.card_type_tree_wdgt, 1, 0)
        self.tag_tree_wdgt = TagsTreeWdgt(\
            component_manager=component_manager, parent=self)
        self.gridLayout.addWidget(self.tag_tree_wdgt, 1, 1)

        criterion = DefaultCriterion(component_manager=self.component_manager)
        for tag in self.database().tags():
            criterion._tag_ids_active.add(tag._id)
        self.display_criterion(criterion)
        self.card_type_tree_wdgt.tree_wdgt.\
            itemChanged.connect(self.criterion_changed)
        self.tag_tree_wdgt.tree_wdgt.\
            itemChanged.connect(self.criterion_changed)
        self.card_type_tree_wdgt.tree_wdgt.\
            itemClicked.connect(self.criterion_clicked)
        self.tag_tree_wdgt.tree_wdgt.\
            itemClicked.connect(self.criterion_clicked)

    def display_criterion(self, criterion):
        self.card_type_tree_wdgt.display(criterion)
        self.tag_tree_wdgt.display(criterion)
        if len(criterion._tag_ids_forbidden):
            self.active_or_forbidden.setCurrentIndex(1)
        else:
            self.active_or_forbidden.setCurrentIndex(0)

    def criterion(self):

        """Build the criterion from the information the user entered in the
        widget.

        """

        criterion = DefaultCriterion(component_manager=self.component_manager)
        criterion = self.card_type_tree_wdgt.checked_to_criterion(criterion)
        # Tag tree contains active tags.
        if self.active_or_forbidden.currentIndex() == 0:
            self.tag_tree_wdgt.checked_to_active_tags_in_criterion(criterion)
        # Tag tree contains forbidden tags.
        else:
            self.tag_tree_wdgt.\
                checked_to_forbidden_tags_in_criterion(criterion)
            for tag in self.database().tags():
                criterion._tag_ids_active.add(tag._id)
        return criterion

    def criterion_clicked(self):
        if self.parent.was_showing_a_saved_set and not self.parent.is_shutting_down:
            self.main_widget().show_information(\
_("Cards you (de)activate now will not be stored in the previously selected set unless you click 'Save this set for later use' again. This allows you to make some quick-and-dirty modifications."))
            self.parent.was_showing_a_saved_set = False

    def criterion_changed(self):
        self.parent.saved_sets.clearSelection()

    def closeEvent(self, event):
        # This allows the state of the tag tree to be saved.
        self.tag_tree_wdgt.close()


#
# configuration.py <Peter.Bienstman@UGent.be>
#

import os

from mnemosyne.libmnemosyne.hook import Hook


class PyQtConfiguration(Hook):

    used_for = "configuration_defaults"

    def run(self):
        for key, value in \
            list({"list_font": None,
             "last_used_card_type_id": "",
             "is_last_used_tags_per_card_type": False,
             "last_used_tags_for_card_type_id": {},
             "last_used_tags": "",
             "sort_column": None,
             "sort_order": None,
             "previous_statistics_page": 0,
             "previous_configuration_wdgt": 0,
             "previous_variant_for_statistics_page": {}, # dict[page] = variant
             "main_window_state": None,
             "add_cards_dlg_state": None,
             "preview_cards_dlg_state": None,
             "edit_card_dlg_state": None,
             "manage_card_types_dlg_state": None,
             "edit_M_sided_card_type_dlg_state": None,
             "plugins_dlg_state": None,
             "sync_dlg_state": None,
             "clone_help_shown": False,
             "browse_cards_dlg_state": None,
             "browse_cards_dlg_splitter_1_state": None,
             "browse_cards_dlg_splitter_2_state": None,
             "browse_cards_dlg_table_settings": None,
             "tag_tree_wdgt_state": None,
             "statistics_dlg_state": None,
             "configuration_dlg_state": None,
             "activate_cards_dlg_state": None,
             "activate_cards_dlg_splitter_state": None,
             "sync_help_shown": False,
             "server_for_sync_as_client": "",
             "port_for_sync_as_client": 8512,
             "username_for_sync_as_client": "",
             "password_for_sync_as_client": "",
             "remember_password_for_sync_as_client": True,
             "started_add_edit_cards_n_times": 0,
             "started_browse_cards_n_times": 0,
             "started_activate_cards_n_times": 0,
             "showed_help_on_renaming_sets": False,
             "showed_help_on_double_clicking_sets": False,
             "showed_help_on_adding_tags": False
            }.items()):
            self.config().setdefault(key, value)



#
# sinhala.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class Sinhala(Language):

    name = _("Sinhala (Sinhalese)")
    used_for = "si"
    feature_description = _("Google translation and text-to-speech.")



#
# plain_text.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.renderer import Renderer


class PlainText(Renderer):

    used_for = None  # All card types.



    def render(self, fact_data, fact_keys, card_type, **render_args):
        text = ""
        for fact_key in fact_keys:
            if fact_key in fact_data and fact_data[fact_key]:
                text += fact_data[fact_key] + "\n"
        return text[:-1]





#
# database.py <Peter.Bienstman@UGent.be>
#

class Database(object):

    """Interface that needs to be implemented by the database object used in the
    Client and the Server.

    It is interesting checking out the implementation in libmnemosyne/databases,
    mainly SQLite_sync.py, as there are some interesting corner cases which
    need to be taken care off in the implementation, e.g. adding and
    immediately deleting an object between two syncs.

    We also provide here an example SQL schema for a minimal openSM2client, i.e.
    one that only supports cards, not facts. It is a stripped-down version of
    libmnemosyne's SQL schema.

    create table cards(
        _id integer primary key,
        id text,
        question text,
        answer text,
        grade integer,
        easiness real,
        acq_reps integer,
        ret_reps integer,
        lapses integer,
        acq_reps_since_lapse integer,
        ret_reps_since_lapse integer,
        last_rep integer,
        next_rep integer
    );
    create index i_cards on cards (id);

    create table tags(
        _id integer primary key,
        id text,
        name text,
    );
    create index i_tags on tags (id);

    create table tags_for_card(
        _card_id integer,
        _tag_id integer
    );
    create index i_tags_for_card on tags_for_card (_card_id);

    /* For object_id, we need to store the full ids as opposed to the _ids.
       When deleting an object, there is no longer a way to get the ids from
       the _ids, and for robustness and interoperability, we need to send the
       ids across when syncing.
    */

    create table log(
        _id integer primary key autoincrement, /* Should never be reused. */
        event_type integer,
        timestamp integer,
        object_id text,
        grade integer,
        easiness real,
        acq_reps integer,
        ret_reps integer,
        lapses integer,
        acq_reps_since_lapse integer,
        ret_reps_since_lapse integer,
        scheduled_interval integer,
        actual_interval integer,
        new_interval integer,
        thinking_time integer,
        last_rep integer,
        next_rep integer,
    );
    create index i_log_timestamp on log (timestamp);
    create index i_log_object_id on log (object_id);

    /* We track the last _id as opposed to the last timestamp, as importing
       another database could add log events with earlier dates, but which
       still need to be synced. Also avoids issues with clock drift. */

    create table partnerships(
        partner text,
        _last_log_id integer
    );

    create table media(
        filename text primary key,
        _hash text
    );

    This minimal client will not be sent any events sister to facts, so it
    does not need to implement them.

    """

    # General info.

    version = "Database version string"

    def path(self):

        """Returns full path of the database."""

        raise NotImplementedError


    def name(self):

        """Returns bare name of the database, without parent paths and
        suffixes.

        """

        raise NotImplementedError

    def media_dir(self):
        raise NotImplementedError

    def user_id(self):
        raise NotImplementedError

    def change_user_id(self, user_id):
        raise NotImplementedError

    # File operations.

    def new(self, path):
        raise NotImplementedError

    def save(self, path=None):
        raise NotImplementedError

    def backup(self):
        raise NotImplementedError

    def restore(self, path):
        raise NotImplementedError

    def load(self, path):
        raise NotImplementedError

    def abandon(self):
        raise NotImplementedError

    def is_empty(self):
        raise NotImplementedError

    # Partnerships.

    def set_sync_partner_info(self, info):
        raise NotImplementedError

    def partners(self):
        raise NotImplementedError

    def create_if_needed_partnership_with(self, partner):
        raise NotImplementedError

    def remove_partnership_with(self, partner):
        raise NotImplementedError

    def merge_partners(self, remote_partners):
        raise NotImplementedError

    def reset_partnerships(self):
        raise NotImplementedError

    def is_sync_reset_needed(self, partner):
        raise NotImplementedError

    def append_to_sync_partner_info(self, partner_info):
        return partner_info

    # Syncing process.

    def update_last_log_index_synced_for(self, partner):
        raise NotImplementedError

    def number_of_log_entries_to_sync_for(self, partner,
            interested_in_old_reps=True):
        raise NotImplementedError

    def number_of_log_entries(self, interested_in_old_reps=True):
        raise NotImplementedError

    def log_entries_to_sync_for(self, partner, interested_in_old_reps=True):
        raise NotImplementedError

    def all_log_entries(self, interested_in_old_reps=True):
        raise NotImplementedError

    def check_for_edited_media_files(self):
        raise NotImplementedError

    def media_filenames_to_sync_for(self, partner):
        raise NotImplementedError

    def all_media_filenames(self):
        raise NotImplementedError

    def apply_log_entry(self, log_entry):
        raise NotImplementedError

    def generate_log_entries_for_settings(self):

        """Needed after binary initial upload/download of the database, to
        ensure that the side effects to config get applied.

        """

        raise NotImplementedError

    # Science log. Only relevant if the client decides to deal with this.
    # Can be left to a libmnemosyne-based server.

    def dump_to_science_log(self):
        pass

    def skip_science_log(self):
        pass



#
# Widget to preview set of sister cards <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component
from mnemosyne.pyqt_ui.review_wdgt import QAOptimalSplit
from mnemosyne.pyqt_ui.ui_preview_cards_dlg import Ui_PreviewCardsDlg


class PreviewCardsDlg(QtWidgets.QDialog, Component, QAOptimalSplit,
                      Ui_PreviewCardsDlg):

    page_up_down_signal = QtCore.pyqtSignal(int)
    UP = 0
    DOWN = 1

    def __init__(self, cards, tag_text, **kwds):

        """We need to provide tag_text explicitly, since it's possible that
        the cards have not yet been added to the database.

        """
        super().__init__(**kwds)
        self.setupUi(self)
        # Needed to make sure we can process e.g. escape keys. Not yet perfect
        # though, as clicking on the html widget eats those keys again.
        self.question.setFocusProxy(self)
        self.answer.setFocusProxy(self)
        QAOptimalSplit.setup(self)
        self.used_for_reviewing = False
        self.setWindowFlags(self.windowFlags() \
            | QtCore.Qt.WindowMinMaxButtonsHint)
        self.setWindowFlags(self.windowFlags() \
            & ~ QtCore.Qt.WindowContextHelpButtonHint)
        self.tag_text = tag_text
        self.cards = cards
        self.index = 0
        state = self.config()["preview_cards_dlg_state"]
        if state:
            self.restoreGeometry(state)
        self.update_dialog()

    def keyPressEvent(self, event):

        """When this dialog is called from the card browser, PageUp and
        PageDown keys can be used to move the previous/next card in the list.

        """

        if event.key() == QtCore.Qt.Key_PageUp:
            self.page_up_down_signal.emit(self.UP)
        elif event.key() == QtCore.Qt.Key_PageDown:
            self.page_up_down_signal.emit(self.DOWN)
        # Note QtWidgets.QWidget.keyPressEvent(self, event) does not seem to work,
        # so we handle the most common keypresses here too.
        if event.key() == QtCore.Qt.Key_Escape:
            self.reject()
        if event.key() in [QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return]:
            self.accept()
        else:
            QtWidgets.QWidget.keyPressEvent(self, event)

    def update_dialog(self):
        self.question_label.setText(_("Question: ") + self.tag_text)
        if len(self.cards) == 1:
            self.previous_button.setVisible(False)
            self.next_button.setVisible(False)
            self.fact_view_name.setVisible(False)
        card = self.cards[self.index]
        stored = self.config()["QA_split"]
        self.config()["QA_split"] = "fixed"
        self.set_question(card.question())
        self.set_answer(card.answer())
        self.reveal_question()
        self.reveal_answer()
        self.config()["QA_split"] = stored
        self.fact_view_name.setText(_(card.fact_view.name) + " (" + \
            str(self.index + 1) + "/" + str(len(self.cards)) + ")")
        self.previous_button.setEnabled(self.index != 0)
        self.next_button.setEnabled(self.index != len(self.cards) - 1)

    def previous(self):
        self.review_widget().stop_media()
        self.index -= 1
        self.update_dialog()

    def next(self):
        self.review_widget().stop_media()
        self.index += 1
        self.update_dialog()

    def _store_state(self):
        self.config()["preview_cards_dlg_state"] = self.saveGeometry()

    def closeEvent(self, event):
        # Generated when clicking the window's close button.
        self.review_widget().stop_media()
        self._store_state()

    def accept(self):
        # 'accept' does not generate a close event.
        self.review_widget().stop_media()
        self._store_state()
        return QtWidgets.QDialog.accept(self)

    def reject(self):
        self.review_widget().stop_media()
        self._store_state()
        QtWidgets.QDialog.reject(self)



#
# log_uploader.py <Peter.Bienstman@UGent.be>
#

import os
import time
import random
import urllib.request, urllib.error, urllib.parse
from threading import Thread

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.component import Component
from mnemosyne.libmnemosyne.utils import traceback_string, MnemosyneError


class LogUploader(Thread, Component):

    def __init__(self, component_manager):
        Thread.__init__(self)
        Component.__init__(self, component_manager)

    def upload(self, filename):

        """Upload a single file to our serverside CGI script.
        Based on code by Jeff Bauer, Aaron Watters, Jim Fulton.

        """

        # Note: we could make the following code a lot cleaner by doing a HTTP
        # PUT, but then we need to change the server side script which would
        # cause problems with backwards compatibility.
        host, port = self.config()["science_server"].split(":")
        uri = '/cgi-bin/cgiupload.py'
        boundary = '%s%s_%s_%s' % \
                    ('-----', int(time.time()), os.getpid(),
                     random.randint(1, 10000))
        f = open(filename, 'rb')
        data = f.read()
        f.close()
        upload_name = str(filename.split("/")[-1].split("\\")[-1])
        hdr, part, total = [], [], []
        hdr.append('Content-Disposition: form-data;' + \
                   ' name="file"; filename="%s"' % (upload_name))
        hdr.append('Content-Type: application/octet-stream')
        hdr.append('Content-Length: %d' % len(data))
        header = (('--%s\n' % boundary) + "\n".join(hdr) + "\n\n")\
            .encode("utf-8")
        footer = ('\n--%s--\n' % boundary).encode("utf-8")
        query = header + data + footer
        contentType = 'multipart/form-data; boundary=%s' % boundary
        contentLength = str(len(query))
        headers = {'Accept' : '*/*',
                   'Proxy-Connection' : 'Keep-Alive',
                   'Content-Type' : contentType,
                   'Content-Length' : contentLength}
        req = urllib.request.Request('http://' + host + ':' + port+uri,
                                     query, headers)
        response = urllib.request.urlopen(req)
        html = str(response.read())
        if "<pre>" in html:
            raise MnemosyneError(html)

    def run(self):
        data_dir = self.config().data_dir
        join = os.path.join
        # Find out which files haven't been uploaded yet.
        dir = os.listdir(join(data_dir, "history"))
        history_files = [x for x in dir if x.endswith(".bz2")]
        uploaded_filename = join(data_dir, "history", "uploaded2")
        if os.path.exists(uploaded_filename):
            upload_log = open(uploaded_filename)
            uploaded = [x.strip() for x in upload_log]
            upload_log.close()
        else:
            uploaded = []
        to_upload = set(history_files) - set(uploaded)
        if len(to_upload) == 0:
            return
        # Upload them to our server.
        upload_log = open(uploaded_filename, 'a')
        try:
            for f in to_upload:
                print(_("Uploading"), f, "...", end=' ')
                filename = join(data_dir, "history", f)
                self.upload(filename)
                print(f, file=upload_log)
                print(_("done!"))
        except Exception as e:
            print(_("Upload failed"))
            print(str(e))
            print(traceback_string())
        upload_log.close()




#
# pashto.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class Pashto(Language):

    name = _("Pashto")
    used_for = "ps"
    feature_description = _("Google translation.")



#
# xml_format.py <Peter.Bienstman@UGent.be>
#

from xml.sax import saxutils
from xml.etree import cElementTree
from openSM2sync.log_entry import LogEntry

PROTOCOL_VERSION = "openSM2sync 1.0   "


class XMLFormat(object):

    """Class handling the conversion from data to XML streams and vice versa.

    Example of typical XML for log entries:

    <openSM2sync number_of_entries='5'>
    <log type='1' o_id='Mnemosyne 2.0-pre posix linux2' time='1268213369'>
    </log>
    <log type='3' o_id='SM2 Mnemosyne' time='1268213369'></log>
    <log n_mem='0' act='0' type='4' sch='0' time='1268213369'></log>
    <log type='6' o_id='068c2472-b1f7-424d-aefa-ae723437702e'
    time='1268213369'><name>abcd</name></log>
    <log n_mem='0' act='0' type='5' sch='0' time='1268213369'></log>
    </openSM2sync>

    Note that the returned XML is a unicode object, and in order to send it
    across a socket e.g., we still need to encode it first.

    """

    mime_type = "text/xml"

    def repr_partner_info(self, info):
        repr_info = "<partner "
        for key, value in list(info.items()):
            if key.lower() == "partners":
                if value:
                    repr_info += "%s=%s " % \
                        (key, saxutils.quoteattr(",".join(value)))
            else:
                if type(value) != str and type(value) != str:
                    value = repr(value)
                repr_info += "%s=%s " % (key, saxutils.quoteattr(value))
        repr_info += "protocol_version=\"%s\"></partner>\n" \
            % (PROTOCOL_VERSION, )
        #import sys; sys.stderr.write(repr_info)
        return repr_info

    def parse_partner_info(self, xml):
        parsed_xml = cElementTree.fromstring(xml)
        partner_info = {}
        partner_info["partners"] = ()
        for key in list(parsed_xml.keys()):
            value = parsed_xml.get(key)
            if value.lower() == "true":
                value = True
            elif value.lower() == "false":
                value = False
            if key.lower() == "partners":
                value = value.split(",")
            partner_info[key] = value
        #import sys; sys.stderr.write(repr(partner_info))
        return partner_info

    # The list of LogEntry keys to be passed on as attributes.
    keys_in_attribs = ["type", "time", "o_id", "sch", "n_mem", "c_time",
        "m_time", "card_t", "fact", "fact_v", "tags", "act", "gr", "e", "l_rp",
        "n_rp", "ac_rp", "rt_rp", "lps", "ac_rp_l", "rt_rp_l", "sch_data",
        "sch_i", "act_i", "th_t"]
    int_keys = ["type", "time", "sch", "n_mem", "act", "c_time", "m_time",
        "act", "gr", "l_rp", "n_rp", "ac_rp", "rt_rp", "lps", "ac_rp_l",
        "rt_rp_l", "sch_data", "sch_i", "act_i", "th_t"]
    float_keys = ["e"]

    def log_entries_header(self, number_of_entries):
        return "<openSM2sync number_of_entries=\"%d\">" % (number_of_entries, )

    def log_entries_footer(self):
        return "</openSM2sync>\n"

    def repr_log_entry(self, log_entry):

        """Converts LogEntry to XML.

        For efficiency reasons we require tag names and attribute values to be
        useable without escaping them with saxutils.quoteattr, so they should
        not contain <, >, &, ... .

        We add an newline after each entry to improve parsing throughput
        (side effect of WSGI 1.x not supporting chunked requests).

        """

        if log_entry is None:
            # Dummy entries for card-based clients.
            return ""
        attribs, tags = "", ""
        for key, value in list(log_entry.items()):
            if key in self.keys_in_attribs:
                attribs += " %s=\"%s\"" % (key, value)
            else:
                # Anki keys can be numbers.
                if key[0].isdigit():
                    key = "___" + key
                tags += "<%s>%s</%s>" % (key, saxutils.escape(value), key)
        xml = "<log%s>%s</log>\n" % (attribs, tags)
        # Strip control characters.
        xml = "".join([i for i in xml if 31 < ord(i) or ord(i) in [9, 10, 13]])
        #import sys; sys.stderr.write(xml)
        return xml

    def parse_log_entries(self, xml):

        """Do incremental parsing of the XML stream.

        See http://effbot.org/zone/element-iterparse.htm

        """
        context = iter(cElementTree.iterparse(xml, events=("start", "end")))
        event, root = next(context)  # 'start' event on openSM2 tag.
        for key, value in list(root.attrib.items()):
            if key == "number_of_entries":
                yield value
        for event, element in context:
            if event == "end" and element.tag == "log":
                log_entry = LogEntry()
                for key, value in list(element.attrib.items()):
                    if key in self.int_keys:
                        value = int(value)
                    elif key in self.float_keys:
                        values = float(value)
                    log_entry[key] = value
                for child in element:
                    if child.tag.startswith("___"):  # Escaped number.
                        child.tag = child.tag[3:]
                    if child.text:
                        log_entry[child.tag] = child.text
                    else:
                        # Should be string instead of None.
                        log_entry[child.tag] = ""
                root.clear()  # Avoid taking up unnecessary memory.
                #import sys; sys.stderr.write(str(log_entry) + "\n")
                yield log_entry

    def repr_message(self, message, traceback=None):
        xml = "<openSM2sync message=\"%s\">" % (message, )
        if traceback:
            xml += "<traceback>%s</traceback>" % saxutils.escape(traceback)
        xml += "</openSM2sync>\n"
        return xml

    def parse_message(self, xml):
        element = cElementTree.XML(xml)
        message = element.attrib["message"]
        traceback = None
        if element.find("traceback") is not None:
            traceback = element.find("traceback").text
        return message, traceback




#
# SQLite_criterion_applier.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.criterion import CriterionApplier
from mnemosyne.libmnemosyne.criteria.default_criterion import DefaultCriterion


class DefaultCriterionApplier(CriterionApplier):

    used_for = DefaultCriterion

    def split_set(self, _set, chunk_size):
        lst = list(_set)
        # Note that [1,2,3][2:666] = [3]
        return [lst[i:i+chunk_size] for i in range(0, len(lst), chunk_size)]

    def set_activity_for_tags_with__id(self, _tag_ids, active):
        if len(_tag_ids) == 0:
            return
        command = """update cards set active=? where _id in
            (select _card_id from tags_for_card where _tag_id in ("""
        args = [1 if active else 0]
        for _tag_id in _tag_ids:
            command += "?,"
            args.append(_tag_id)
        command = command.rsplit(",", 1)[0] + "))"
        self.database().con.execute(command, args)

    def apply_to_database(self, criterion):
        if len(criterion._tag_ids_forbidden) != 0:
            assert len(criterion._tag_ids_active) != 0
        db = self.database()
        # If every tag is active, take a shortcut.
        tag_count = db.con.execute("select count() from tags").fetchone()[0]
        if len(criterion._tag_ids_active) == tag_count:
            db.con.execute("update cards set active=1")
        else:
            # Turn off everything.
            db.con.execute("update cards set active=0")
            # Turn on active tags. Limit to 500 at a time to deal with
            # SQLite limitations.
            for chunked__tag_ids in self.split_set(\
                    criterion._tag_ids_active, 500):
                self.set_activity_for_tags_with__id(chunked__tag_ids, active=1)
        # Turn off inactive card types and views.
        command = "update cards set active=0 where "
        args = []
        for card_type_id, fact_view_id in \
                criterion.deactivated_card_type_fact_view_ids:
            command += "(cards.fact_view_id=? and cards.card_type_id=?)"
            command += " or "
            args.append(fact_view_id)
            args.append(card_type_id)
        command = command.rsplit("or ", 1)[0]
        if criterion.deactivated_card_type_fact_view_ids:
            db.con.execute(command, args)
        # Turn off forbidden tags. Limit to 500 at a time to deal with
        # SQLite limitations.
        for chunked__tag_ids in self.split_set(\
                criterion._tag_ids_forbidden, 500):
            self.set_activity_for_tags_with__id(chunked__tag_ids, active=0)



#
# dlgs.py <Peter.Bienstman@UGent.be>
#

import _dlgs as _
import mnemosyne.libmnemosyne.ui_components.dialogs as dialogs


class AddCardsDlg(dialogs.AddCardsDialog):

    def activate(self):
        _.add_cards_dlg_activate()


class EditCardDlg(dialogs.EditCardDialog):
    
    def __init__(self, card, component_manager, allow_cancel=True):
        self.card = card
        self.allow_cancel = allow_cancel

    def activate(self):
        _.edit_card_dlg_activate(self.card.id, self.allow_cancel)

        
class ActivateCardsDlg(dialogs.ActivateCardsDialog):

    def activate(self):
        _.activate_cards_dlg_activate()
        
    
class BrowseCardsDlg(dialogs.BrowseCardsDialog):
    
    def activate(self):
        _.browse_cards_dlg_activate()


class CardAppearanceDlg(dialogs.CardAppearanceDialog):
    
    def activate(self):
        _.card_appearance_dlg_activate()
        
    
class ActivatePluginsDlg(dialogs.ActivatePluginsDialog):
    
    def activate(self):
        _.activate_plugins_dlg_activate()


class ManageCardTypesDlg(dialogs.ManageCardTypesDialog):
    
    def activate(self):
        _.manage_card_types_dlg_activate()


class StatisticsDlg(dialogs.StatisticsDialog):

    def activate(self):
        _.statistics_dlg_activate()


class ConfigurationDlg(dialogs.ConfigurationDialog):

    def activate(self):
        _.configuration_dlg_activate()


class SyncDlg(dialogs.SyncDialog):
    
    def activate(self):
        _.sync_dlg_activate()



#
# web_server.py <Peter.Bienstman@UGent.be>
#

import os
import cgi
import sys
import time
import locale
import http.client
import threading

from mnemosyne.libmnemosyne import Mnemosyne
from mnemosyne.libmnemosyne.utils import localhost_IP
from mnemosyne.libmnemosyne.component import Component


class ReleaseDatabaseAfterTimeout(threading.Thread):

    def __init__(self, port):
        threading.Thread.__init__(self)
        self.port = port
        self.ping()

    def ping(self):
        self.last_ping = time.time()

    def run(self):
        while time.time() < self.last_ping + 2*60:
            time.sleep(1)
        con = http.client.HTTPConnection("localhost", self.port)
        con.request("GET", "/release_database")
        response = con.getresponse()


class StopServerAfterTimeout(threading.Thread):

    """Stop server after a certain timeout, so that it has
    enough time to serve the final page.

    """

    def __init__(self, wsgi_server):
        threading.Thread.__init__(self)
        self.wsgi_server = wsgi_server

    def run(self):
        time.sleep(3)
        self.wsgi_server.stop()


class WebServer(Component):

    def __init__(self, port, data_dir, config_dir, filename, **kwds):
        if "client_on_same_machine_as_server" in kwds:
            self.client_on_same_machine_as_server = \
                kwds["client_on_same_machine_as_server"]
            del kwds["client_on_same_machine_as_server"]
        else:
            self.client_on_same_machine_as_server = False
        super().__init__(**kwds)
        self.wsgi_server = None
        self.port = port
        self.data_dir = data_dir
        self.config_dir = config_dir
        self.filename = filename
        # When restarting the server, make sure we discard info from the
        # browser resending the form from the previous session.
        self.is_just_started = True
        self.is_mnemosyne_loaded = False
        self.is_shutting_down = False

    def activate(self):
        Component.activate(self)
        # Late import to speed up application startup.
        from cheroot import wsgi
        self.wsgi_server = wsgi.Server(\
            ("0.0.0.0", self.port), self.wsgi_app, server_name="localhost",
            numthreads=1, timeout=5)
        # We need to set the timeout relatively low, otherwise it will take
        # too long for the server to process a 'stop' request.

    def serve_until_stopped(self):
        try:
            self.wsgi_server.start() # Sets self.wsgi_server.ready
        except KeyboardInterrupt:
            self.wsgi_server.stop()
            self.unload_mnemosyne()

    def stop(self):
        if self.wsgi_server:
            self.wsgi_server.stop()
        self.unload_mnemosyne()

    def load_mnemosyne(self):
        self.mnemosyne = Mnemosyne(upload_science_logs=True,
            interested_in_old_reps=True)
        self.mnemosyne.components.insert(0, (
            ("mnemosyne.libmnemosyne.gui_translators.gettext_gui_translator",
             "GetTextGuiTranslator")))
        self.mnemosyne.components.append(\
            ("mnemosyne.libmnemosyne.ui_components.main_widget",
             "MainWidget"))
        self.mnemosyne.components.append(\
            ("mnemosyne.web_server.web_server_render_chain",
             "WebServerRenderChain"))
        self.mnemosyne.gui_for_component["ScheduledForgottenNew"] = [\
            ("mnemosyne.web_server.review_wdgt",
             "ReviewWdgt")]
        self.mnemosyne.gui_for_component["NewOnly"] = [\
            ("mnemosyne.web_server.review_wdgt",
             "ReviewWdgt")]
        self.mnemosyne.gui_for_component["CramAll"] = [\
            ("mnemosyne.web_server.review_wdgt",
             "ReviewWdgt")]
        self.mnemosyne.gui_for_component["CramRecent"] = [\
            ("mnemosyne.web_server.review_wdgt",
             "ReviewWdgt")]
        self.mnemosyne.initialise(self.data_dir, config_dir=self.config_dir,
            filename=self.filename, automatic_upgrades=False)
        self.save_after_n_reps = self.mnemosyne.config()["save_after_n_reps"]
        self.mnemosyne.config()["save_after_n_reps"] = 1
        self.mnemosyne.config()["study_mode"] = "ScheduledForgottenNew"
        self.mnemosyne.config()["QA_split"] = "fixed"
        self.mnemosyne.review_widget().set_client_on_same_machine_as_server(\
            self.client_on_same_machine_as_server)
        self.mnemosyne.controller().reset_study_mode()
        self.is_mnemosyne_loaded = True
        self.release_database_after_timeout = \
            ReleaseDatabaseAfterTimeout(self.port)
        self.release_database_after_timeout.start()

    def unload_mnemosyne(self):
        if not self.is_mnemosyne_loaded:
            return
        self.mnemosyne.config()["save_after_n_reps"] = self.save_after_n_reps
        self.mnemosyne.finalise()
        self.is_mnemosyne_loaded = False

    def wsgi_app(self, environ, start_response):
        filename = environ["PATH_INFO"]
        if filename == "/status":
            response_headers = [("Content-type", "text/html")]
            start_response("200 OK", response_headers)
            return [b"200 OK"]
        # Sometimes, even after the user has clicked 'exit' in the page,
        # a browser sends a request for e.g. an audio file.
        if self.is_shutting_down and filename != "/release_database":
            response_headers = [("Content-type", "text/html")]
            start_response("503 Service Unavailable", response_headers)
            return [b"Server stopped"]
        # Load database if needed.
        if not self.is_mnemosyne_loaded and filename != "/release_database":
            self.load_mnemosyne()
        if self.is_mnemosyne_loaded:
            self.release_database_after_timeout.ping()
        # All our request return to the root page, so if the path is '/',
        # return the html of the review widget.
        if filename == "/":
            # Process clicked buttons in the form.
            form = cgi.FieldStorage(fp=environ["wsgi.input"], environ=environ)
            if "show_answer" in form and not self.is_just_started:
                self.mnemosyne.review_widget().show_answer()
                page = self.mnemosyne.review_widget().to_html()
            elif "grade" in form and not self.is_just_started:
                grade = int(form["grade"].value)
                self.mnemosyne.review_widget().grade_answer(grade)
                page = self.mnemosyne.review_widget().to_html()
            elif "star" in form:
                self.mnemosyne.controller().star_current_card()
                page = self.mnemosyne.review_widget().to_html()
            elif "exit" in form:
                self.unload_mnemosyne()
                page = "Server stopped"
                self.wsgi_server.stop()
                self.stop_server_after_timeout = \
                    StopServerAfterTimeout(self.wsgi_server)
                self.stop_server_after_timeout.start()
                self.is_shutting_down = True
            else:
                page = self.mnemosyne.review_widget().to_html()
            if self.is_just_started:
                self.is_just_started = False
            # Serve the web page.
            response_headers = [("Content-type", "text/html")]
            start_response("200 OK", response_headers)
            return [page]
        elif filename == "/api/scheduled":
            scheduled_count, _, _ = \
                self.mnemosyne.review_widget().review_controller().counters()

            response_headers = [("Content-type", "text/plain")]
            start_response("200 OK", response_headers)

            return [str(scheduled_count).encode(encoding='UTF-8')]
        elif filename == "/api/non_memorized":
            _, non_memorised_count, _ = \
                self.mnemosyne.review_widget().review_controller().counters()

            response_headers = [("Content-type", "text/plain")]
            start_response("200 OK", response_headers)

            return [str(non_memorised_count).encode(encoding='UTF-8')]
        elif filename == "/api/active":
            _, _, active_count = \
                self.mnemosyne.review_widget().review_controller().counters()

            response_headers = [("Content-type", "text/plain")]
            start_response("200 OK", response_headers)

            return [str(active_count).encode(encoding='UTF-8')]
        elif filename == "/release_database":
            self.unload_mnemosyne()
            response_headers = [("Content-type", "text/html")]
            start_response("200 OK", response_headers)
            return [b"200 OK"]
        # We need to serve a media file.
        else:
            # Late import to speed up application startup.
            from webob import Request
            from webob.static import FileApp
            full_path = self.mnemosyne.database().media_dir()
            for word in filename.split("/"):
                full_path = os.path.join(full_path, word)
            request = Request(environ)
            if os.path.exists(full_path):
                etag = "%s-%s-%s" % (os.path.getmtime(full_path),
                    os.path.getsize(full_path), hash(full_path))
            else:
                etag = "none"
            app = FileApp(full_path, etag=etag)
            return app(request)(environ, start_response)


class WebServerThread(threading.Thread, WebServer):

    """Basic threading implementation of the sync server, suitable for text-
    based UIs. A GUI-based client will want to override several functions
    in Server and ServerThread in view of the interaction between multiple
    threads and the GUI event loop.

    """

    def __init__(self, component_manager, client_on_same_machine_as_server=False):
        self.client_on_same_machine_as_server = client_on_same_machine_as_server
        threading.Thread.__init__(self)
        self.config = component_manager.current("config")
        WebServer.__init__(self,
            self.config["web_server_port"], self.config.data_dir,
            self.config.config_dir, self.config["last_database"],
            component_manager=component_manager,
            client_on_same_machine_as_server=self.client_on_same_machine_as_server)

    def run(self):
        self.activate()
        if not self.client_on_same_machine_as_server:  # Could fail if we are offline.
            print(("Web server listening on http://" + \
                localhost_IP() + ":" + str(self.config["web_server_port"])))
        self.serve_until_stopped()



#
# kurdish.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.libmnemosyne.language import Language


class Kurdish(Language):

    name = _("Kurdish")
    used_for = "ku"
    feature_description = _("Google translation.")



#
# cramming.py <Peter.Bienstman@UGent.be>
#

from mnemosyne.libmnemosyne.schedulers.SM2_mnemosyne import SM2Mnemosyne

RANDOM = 0
EARLIEST_FIRST = 1
LATEST_FIRST = 2
MOST_LAPSES_FIRST = 3


class Cramming(SM2Mnemosyne):

    UNSEEN = 0
    WRONG = 1
    RIGHT = 2

    name = "cramming"

    def __init__(self, component_manager):
        SM2Mnemosyne.__init__(self, component_manager)
        self.first_run = True

    def reset(self, new_only=False):
        if self.first_run and self.config()["cramming_store_state"] == False:
            self.database().set_scheduler_data(self.UNSEEN)
        self.first_run = False
        SM2Mnemosyne.reset(self, new_only)
        self.stage = 1

    def rebuild_queue(self, learn_ahead=False):
        db = self.database()
        if not db.is_loaded() or not db.active_count():
            return
        max_ret_reps = self.config()["max_ret_reps_for_recent_cards"] \
            if self.new_only else -1
        if self.new_only and db.recently_memorised_count(max_ret_reps) == 0:
            return
        self._card_ids_in_queue = []
        self._fact_ids_in_queue = []
        self.criterion = db.current_criterion()
        # Determine sort key.
        if self.config()["cramming_order"] == RANDOM:
            sort_key = "random"
        elif self.config()["cramming_order"] == EARLIEST_FIRST:
            sort_key = "next_rep"
        elif self.config()["cramming_order"] == LATEST_FIRST:
            sort_key = "next_rep desc"
        elif self.config()["cramming_order"] == MOST_LAPSES_FIRST:
            sort_key = "lapses desc"
        # Stage 1: do all the unseen cards.
        if self.stage == 1:
            for _card_id, _fact_id in db.cards_with_scheduler_data(self.UNSEEN,
                    sort_key=sort_key, limit=25, max_ret_reps=max_ret_reps):
                if _fact_id not in self._fact_ids_in_queue:
                    self._card_ids_in_queue.append(_card_id)
                    self._fact_ids_in_queue.append(_fact_id)
            if len(self._card_ids_in_queue):
                return
            self.stage = 2
        # Stage 2: do the cards we got wrong.
        if self.stage == 2:
            for _card_id, _fact_id in db.cards_with_scheduler_data(self.WRONG,
                    sort_key=sort_key, limit=25, max_ret_reps=max_ret_reps):
                if _fact_id not in self._fact_ids_in_queue:
                    self._card_ids_in_queue.append(_card_id)
                    self._fact_ids_in_queue.append(_fact_id)
            if len(self._card_ids_in_queue):
                return
        # Start again.
        self.database().set_scheduler_data(self.UNSEEN)
        self.stage = 1
        self.rebuild_queue()

    def grade_answer(self, card, new_grade, dry_run=False):
        # The dry run mode is typically used to determine the intervals
        # for the different grades, so we don't want any side effects
        # from hooks running then.
        if not dry_run:
            for f in self.component_manager.all("hook", "before_repetition"):
                f.run(card)
        # Do the actual grading.
        if new_grade <= 1:
            card.scheduler_data = self.WRONG
        else:
            card.scheduler_data = self.RIGHT
        # Run hooks.
        self.criterion.apply_to_card(card)
        for f in self.component_manager.all("hook", "after_repetition"):
            f.run(card)
        return 0


#
# review_wdgt_cramming.py <Peter.Bienstman@UGent.be>
#

from PyQt5 import QtCore, QtGui, QtWidgets

from mnemosyne.libmnemosyne.gui_translator import _
from mnemosyne.pyqt_ui.review_wdgt import ReviewWdgt


class ReviewWdgtCramming(ReviewWdgt):

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self.grade_0_button.setText(_("Wrong"))
        self.grade_1_button.hide()
        self.line.hide()
        self.grade_2_button.hide()
        self.grade_3_button.hide()
        self.grade_4_button.hide()
        self.grade_5_button.setText(_("Right"))
        self.grade_5_button.setFocus()
        parent = self.parent()
        self.wrong = QtWidgets.QLabel("", parent.status_bar)
        self.unseen = QtWidgets.QLabel("", parent.status_bar)
        self.active = QtWidgets.QLabel("", parent.status_bar)
        parent.clear_status_bar()
        parent.add_to_status_bar(self.wrong)
        parent.add_to_status_bar(self.unseen)
        parent.add_to_status_bar(self.active)

    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.LanguageChange:
            self.retranslateUi(self)
        # Upon start, there will be a change event before the grade
        # buttons have been created.
        if hasattr(self, "grade_0_button"):
            self.grade_0_button.setText(_("Wrong"))
            self.grade_5_button.setText(_("Right"))
        QtWidgets.QWidget.changeEvent(self, event)

    def keyPressEvent(self, event):
        if self.review_controller().is_answer_showing():
            if event.key() in [QtCore.Qt.Key_0, QtCore.Qt.Key_1]:
                return self.grade_answer(0)
            elif event.key() in [QtCore.Qt.Key_2, QtCore.Qt.Key_3,
                QtCore.Qt.Key_4, QtCore.Qt.Key_5]:
                return self.grade_answer(5)
        super().keyPressEvent(event)

    def update_status_bar_counters(self):
        wrong_count, unseen_count, active_count = \
                   self.review_controller().counters()
        self.wrong.setText(_("Wrong:") + " %d " % wrong_count)
        self.unseen.setText(_("Unseen:") + " %d " % unseen_count)
        self.active.setText(_("Active:") + " %d " % active_count)




#
# partner.py <Peter.Bienstman@UGent.be>
#

import os
if "ANDROID" in os.environ:
    from mnemosyne.android_python.utf8_filenames import *
from openSM2sync.utils import normalise_path


class Partner(object):

    """Common code between Client and Server."""

    def __init__(self, ui):
        self.ui = ui

    def stream_binary_file(self, filename, progress_bar=True):
        filename = normalise_path(filename)
        binary_file = open(filename, "rb")
        file_size = os.path.getsize(filename)
        buffer = binary_file.read(self.BUFFER_SIZE)
        if progress_bar:
            self.ui.set_progress_range(file_size)
            self.ui.set_progress_update_interval(file_size/50)
            self.ui.increase_progress(len(buffer))
        while buffer:
            yield buffer
            buffer = binary_file.read(self.BUFFER_SIZE)
            if progress_bar:
                self.ui.increase_progress(len(buffer))
        if progress_bar:
            self.ui.set_progress_value(file_size)

    def download_binary_file(self, stream, filename, file_size,
                             progress_bar=True):
        filename = normalise_path(filename)
        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
            os.makedirs(directory)
        downloaded_file = open(filename, "wb")
        if progress_bar:
            self.ui.set_progress_range(file_size)
            self.ui.set_progress_update_interval(file_size/50)
        buffer = stream.read(self.BUFFER_SIZE)
        while buffer:
            downloaded_file.write(buffer)
            if progress_bar:
                self.ui.increase_progress(len(buffer))
            buffer = stream.read(self.BUFFER_SIZE)
        if progress_bar:
            self.ui.set_progress_value(file_size)
        downloaded_file.close()




from .template import Template
from .view import View

def render(template, context=None, **kwargs):
    context = context and context.copy() or {}
    context.update(kwargs)
    return Template(template, context).render()
